# coding=utf-8

import logging
import mimetypes
import os
import tempfile
import threading
from typing import Tuple, IO, Optional, TYPE_CHECKING

import magic
import telegram
import time
from moviepy.video.io.VideoFileClip import VideoFileClip
from telegram.ext import MessageHandler, Filters
from PIL import Image
from telegram.utils.helpers import escape_markdown

from ehforwarderbot import EFBChat, EFBMsg, coordinator
from ehforwarderbot.constants import MsgType
from ehforwarderbot.exceptions import EFBMessageTypeNotSupported, EFBChatNotFound, \
    EFBMessageError, EFBMessageNotFound, EFBOperationNotSupported
from ehforwarderbot.message import EFBMsgLocationAttribute
from ehforwarderbot.status import EFBMessageRemoval
from . import utils
from .msg_type import get_msg_type, TGMsgType
from .locale_mixin import LocaleMixin

if TYPE_CHECKING:
    from . import TelegramChannel
    from .bot_manager import TelegramBotManager
    from .db import DatabaseManager


class MasterMessageProcessor(LocaleMixin):
    """
    Processes messages from Telegram user and delivers to the slave channels
    """

    DELETE_FLAG = 'rm`'
    FAIL_FLAG = '__fail__'

    # Constants
    TYPE_DICT = {
        TGMsgType.Text: MsgType.Text,
        TGMsgType.Audio: MsgType.Audio,
        TGMsgType.Document: MsgType.File,
        TGMsgType.Photo: MsgType.Image,
        TGMsgType.Sticker: MsgType.Sticker,
        TGMsgType.Video: MsgType.Video,
        TGMsgType.Voice: MsgType.Audio,
        TGMsgType.Location: MsgType.Location,
        TGMsgType.Venue: MsgType.Location,
    }

    def __init__(self, channel: 'TelegramChannel'):
        self.channel: 'TelegramChannel' = channel
        self.bot: 'TelegramBotManager' = channel.bot_manager
        self.db: 'DatabaseManager' = channel.db
        self.bot.dispatcher.add_handler(MessageHandler(
            Filters.text | Filters.photo | Filters.sticker | Filters.document |
            Filters.venue | Filters.location | Filters.audio | Filters.voice | Filters.video,
            self.msg_thread_creator, edited_updates=True
        ))
        self.logger: logging.Logger = logging.getLogger(__name__)

        self.channel_id: str = self.channel.channel_id

    def msg_thread_creator(self, bot, update):
        """Process message in a thread, to ensure it doesn't block the main thread."""
        threading.Thread(target=self.msg, args=(bot, update)).run()

    def msg(self, bot, update: telegram.Update):
        """
        Process, wrap and dispatch messages from user.

        Args:
            bot: Telegram Bot instance
            update: Message update
        """

        message: telegram.Message = update.message or update.edited_message or \
                                    update.channel_post or update.edited_channel_post

        self.logger.debug("Received message from Telegram: %s", message.to_dict())
        multi_slaves = False

        if message.chat.id != message.from_user.id:  # from group
            assocs = self.db.get_chat_assoc(master_uid=utils.chat_id_to_str(self.channel_id, message.chat.id))
            if len(assocs) > 1:
                multi_slaves = True

        reply_to = bool(getattr(message, "reply_to_message", None))
        private_chat = message.chat.id == message.from_user.id

        if (private_chat or multi_slaves) and not reply_to:
            candidates = self.db.get_recent_slave_chats(message.chat.id) or \
                         self.db.get_chat_assoc(master_uid=utils.chat_id_to_str(self.channel_id, message.chat.id))[:5]
            if candidates:
                tg_err_msg = message.reply_text(self._("Error: No recipient specified.\n"
                                                       "Please reply to a previous message. (MS01)"), quote=True)
                self.channel.chat_binding.register_suggestions(update, candidates,
                                                               update.effective_chat.id, tg_err_msg.message_id)

            else:
                message.reply_text(self._("Error: No recipient specified.\n"
                                          "Please reply to a previous message. (MS02)"), quote=True)
        else:
            return self.process_telegram_message(bot, update)

    def process_telegram_message(self, bot: telegram.Bot,
                                 update: telegram.Update,
                                 channel_id: Optional[str] = None,
                                 chat_id: Optional[str] = None,
                                 target_msg: Optional[str] = None):
        """
        Process messages came from Telegram.

        Args:
            bot: Telegram bot
            update: Telegram message update
            channel_id: Slave channel ID if specified
            chat_id: Slave chat ID if specified
            target_msg: Target slave message if specified

        Returns:

        """
        target: str = None
        target_channel: str = None
        target_log: self.db.MsgLog = None
        # Message ID for logging
        message_id = utils.message_id_to_str(update=update)

        multi_slaves: bool = False
        destination: str = None
        slave_msg: EFBMsg = None

        message: telegram.Message = update.effective_message

        edited = bool(update.edited_message or update.edited_channel_post)
        self.logger.debug('[%s] Message is edited: %s, %s',
                          message_id, edited, message.edit_date)

        private_chat = update.effective_chat.type == telegram.Chat.PRIVATE

        if not private_chat:  # from group
            linked_chats = self.db.get_chat_assoc(master_uid=utils.chat_id_to_str(
                self.channel_id, update.effective_chat.id))
            if len(linked_chats) == 1:
                destination = linked_chats[0]
            elif len(linked_chats) > 1:
                multi_slaves = True

        reply_to = bool(getattr(message, "reply_to_message", None))

        # Process predefined target (slave) chat.
        if channel_id and chat_id:
            destination = utils.chat_id_to_str(channel_id, chat_id)
            if target_msg:
                target_log = self.db.get_msg_log(master_msg_id=target_msg)
                if target_log:
                    target = target_log.slave_origin_uid
                    target_channel, target_uid = utils.chat_id_str_to_id(target)
                else:
                    self.logger.info("[%s], Predefined chat %d.%d with target msg")
                    return self.bot.reply_error(update,
                                                self._("Message is not found in database. "
                                                "Please try with another message. (UC07)"))
        elif private_chat:
            if reply_to:
                destination = self.db.get_msg_log(master_msg_id=utils.message_id_to_str(
                    message.reply_to_message.chat.id,
                    message.reply_to_message.message_id))
                if destination:
                    destination = destination.slave_origin_uid
                else:
                    return self.bot.reply_error(update,
                                                self._("Message is not found in database. "
                                                "Please try with another one. (UC03)"))
            else:
                return self.bot.reply_error(update,
                                            self._("Please reply to an incoming message. (UC04)"))
        else:  # group chat
            if multi_slaves:
                if reply_to:
                    destination = self.db.get_msg_log(master_msg_id=utils.message_id_to_str(
                        message.reply_to_message.chat.id,
                        message.reply_to_message.message_id))
                    if destination:
                        destination = destination.slave_origin_uid
                    else:
                        return self.bot.reply_error(update,
                                                    self._("Message is not found in database. "
                                                    "Please try with another one. (UC05)"))
                else:
                    return self.bot.reply_error(update,
                                                self._("This group is linked to multiple remote chats. "
                                                "Please reply to an incoming message. "
                                                "To unlink all remote chats, please send /unlink_all . (UC06)"))
            elif destination:
                if reply_to:
                    target_log: self.db.MsgLog = \
                        self.db.get_msg_log(master_msg_id=utils.message_id_to_str(
                                                               message.reply_to_message.chat.id,
                                                               message.reply_to_message.message_id))
                    if target_log:
                        target = target_log.slave_origin_uid
                        target_channel, target_uid = utils.chat_id_str_to_id(target)
                    else:
                        return self.bot.reply_error(update,
                                                    self._("Message is not found in database. "
                                                    "Please try with another message. (UC07)"))
            else:
                return self.bot.reply_error(update,
                                            self._("This group is not linked to any chat. (UC06)"))

        self.logger.debug("[%s] Telegram received. From private chat: %s; Group has multiple linked chats: %s; "
                          "Message replied to another message: %s", message_id, private_chat, multi_slaves, reply_to)
        self.logger.debug("[%s] Destination chat = %s", message_id, destination)
        channel, uid = utils.chat_id_str_to_id(destination)
        if channel not in coordinator.slaves:
            return self.bot.reply_error(update, self._("Internal error: Channel \"{0}\" not found.").format(channel))

        m = EFBMsg()
        try:
            m.uid = message_id
            mtype = get_msg_type(message)
            # Chat and author related stuff
            m.author = EFBChat(self.channel).self()
            m.chat = EFBChat(coordinator.slaves[channel])
            m.chat.chat_uid = uid
            m.deliver_to = coordinator.slaves[channel]
            if target and target_channel == channel:
                trgt_msg = EFBMsg()
                trgt_msg.type = MsgType.Text
                trgt_msg.text = target_log.text
                trgt_msg.uid = target_log.slave_message_id
                trgt_msg.chat = EFBChat(coordinator.slaves[target_channel])
                trgt_msg.chat.chat_name = target_log.slave_origin_display_name
                trgt_msg.chat.chat_alias = target_log.slave_origin_display_name
                trgt_msg.chat.chat_uid = utils.chat_id_str_to_id(target_log.slave_origin_uid)[1]
                if target_log.slave_member_uid:
                    trgt_msg.author = EFBChat(coordinator.slaves[target_channel])
                    trgt_msg.author.chat_name = target_log.slave_member_display_name
                    trgt_msg.author.chat_alias = target_log.slave_member_display_name
                    trgt_msg.author.chat_uid = target_log.slave_member_uid
                elif target_log.sent_to == 'master':
                    trgt_msg.author = trgt_msg.chat
                else:
                    trgt_msg.author = EFBChat(self).self()
                m.target = trgt_msg

                self.logger.debug("[%s] This message replies to another message of the same channel.\n"
                                  "Chat ID: %s; Message ID: %s.", message_id, trgt_msg.chat.chat_uid, trgt_msg.uid)
            # Type specific stuff
            self.logger.debug("[%s] Message type from Telegram: %s", message_id, mtype)

            if self.TYPE_DICT.get(mtype, None):
                m.type = self.TYPE_DICT[mtype]
                self.logger.debug("[%s] EFB message type: %s", message_id, mtype)
            else:
                self.logger.info("[%s] Message type %s is not supported by ETM", message_id, mtype)
                raise EFBMessageTypeNotSupported("Message type %s is not supported by ETM" % mtype)

            if m.type not in coordinator.slaves[channel].supported_message_types:
                self.logger.info("[%s] Message type %s is not supported by channel %s",
                                 message_id, m.type.name, channel)
                raise EFBMessageTypeNotSupported("Message type %s is not supported by channel %s" % (
                    m.type, coordinator.slaves[channel].channel_name
                ))

            # Flag for edited message
            msg_md_text = message.text and message.text_markdown_urled
            if msg_md_text == escape_markdown(message.text):
                msg_md_text = message.text
            if edited:
                m.edit = True
                text = msg_md_text or message.caption
                msg_log = self.db.get_msg_log(master_msg_id=utils.message_id_to_str(update=update))
                if not msg_log or msg_log == self.FAIL_FLAG:
                    raise EFBMessageNotFound()
                m.uid = msg_log.slave_message_id
                if text.startswith(self.DELETE_FLAG):
                    coordinator.send_status(EFBMessageRemoval(
                        source_channel=self.channel,
                        destination_channel=coordinator.slaves[channel],
                        message=m
                    ))
                    self.db.delete_msg_log(master_msg_id=utils.message_id_to_str(update=update))
                    m = None
                    return
                self.logger.debug('[%s] Message is edited (%s)', m.uid, m.edit)

            # Enclose message as an EFBMsg object by message type.
            if mtype == TGMsgType.Text:
                m.text = msg_md_text
            elif mtype == TGMsgType.Photo:
                m.text = message.caption
                m.file, m.mime, m.filename, m.path = self._download_file(message.photo[-1], None)
            elif mtype == TGMsgType.Sticker:
                # Convert WebP to the more common PNG
                m.text = ""
                m.file, m.mime, m.filename, m.path = self._download_file(message.sticker, 'image/webp')
                self.logger.debug("[%s] Trying to convert WebP sticker (%s) to PNG.", message_id, m.path)
                f = tempfile.NamedTemporaryFile(suffix=".png")
                Image.open(m.file).convert("RGBA").save(f, 'png')
                m.file.close()
                m.file, m.mime, m.filename, m.path = f, 'image/png', os.path.basename(f.name), f.name
                self.logger.debug("[%s] WebP sticker is converted to PNG (%s).", message_id, f.name)
            elif mtype == TGMsgType.Document:
                m.text = message.caption
                self.logger.debug("[%s] Telegram message type is document.", message_id)
                m.filename = getattr(message.document, "file_name", None) or None
                if message.document.mime_type == "video/mp4":
                    self.logger.debug("[%s] Telegram message is a \"Telegram GIF\".", message_id)
                    m.type = MsgType.Image
                    m.file, m.mime, m.filename, m.path = self._download_gif(message.document)
                else:
                    m.type = MsgType.File
                    m.file, m.mime, filename, m.path = self._download_file(message.document,
                                                                           message.document.mime_type)
                    m.filename = m.filename or filename
                m.mime = message.document.mime_type or m.mime
            elif mtype == TGMsgType.Video:
                m.type = MsgType.Video
                m.text = message.caption
                m.file, m.mime, m.filename, m.path = self._download_file(message.video,
                                                                         message.video.mime_type)
            elif mtype == TGMsgType.Audio:
                m.type = MsgType.Audio
                m.text = "%s - %s\n%s" % (
                    message.audio.title, message.audio.performer, message.caption)
                m.file, m.mime, m.filename, m.path = self._download_file(message.audio,
                                                                         message.audio.mime_type)
            elif mtype == TGMsgType.Voice:
                m.type = MsgType.Audio
                m.text = message.caption
                m.file, m.mime, m.filename, m.path = self._download_file(message.voice,
                                                                         message.voice.mime_type)
            elif mtype == TGMsgType.Location:
                m.type = MsgType.Location
                m.text = "Location"
                m.attributes = EFBMsgLocationAttribute(
                    message.location.latitude,
                    message.location.longitude
                )
            elif mtype == TGMsgType.Venue:
                m.type = MsgType.Location
                m.text = message.location.title + "\n" + message.location.adderss
                m.attributes = EFBMsgLocationAttribute(
                    message.venue.location.latitude,
                    message.venue.location.longitude
                )
            else:
                raise EFBMessageTypeNotSupported(self._("Message type {0} is not supported.").format(mtype))
                # return self.bot.reply_error(update, "Message type not supported. (MN02)")

            slave_msg = coordinator.send_message(m)
        except EFBChatNotFound as e:
            self.bot.reply_error(update, e.args[0] or self._("Chat is not found."))
        except EFBMessageTypeNotSupported as e:
            self.bot.reply_error(update, e.args[0] or self._("Message type is not supported."))
        except EFBOperationNotSupported as e:
            self.bot.reply_error(update, self._("Message editing is not supported.\n\n{!r}".format(e)))
        except EFBMessageError as e:
            self.bot.reply_error(update, self._("Message is not sent.\n\n{!r}".format(e)))
        finally:
            if m:
                msg_log_d = {
                    "master_msg_id": utils.message_id_to_str(update=update),
                    "text": m.text or "Sent a %s" % m.type,
                    "slave_origin_uid": utils.chat_id_to_str(chat=m.chat),
                    "slave_origin_display_name": "__chat__",
                    "msg_type": m.type,
                    "sent_to": "slave",
                    "slave_message_id": None if m.edit else "%s.%s" % (self.FAIL_FLAG, int(time.time())),
                    "update": m.edit
                }
                if slave_msg:
                    msg_log_d['slave_message_id'] = slave_msg.uid
                self.db.add_msg_log(**msg_log_d)
                if m.file:
                    m.file.close()

    def _download_file(self, file_obj: telegram.File, mime: str) -> Tuple[IO[bytes], str, str, str]:
        """
        Download media file from telegram platform.

        Args:
            file_obj (telegram.File): PTB file object
            mime (str): MIME type of the message

        Returns:
            Tuple[IO[bytes], str, str, str]:
                ``tempfile`` file-like object, MIME type, proposed file name, file path

        Raises:
            EFBMessageError: When file exceeds the maximum download size.
        """
        size = getattr(file_obj, "file_size", None)
        file_id = file_obj.file_id
        if size and size > telegram.constants.MAX_FILESIZE_DOWNLOAD:
            raise EFBMessageError(self._("Attachment is too large. Maximum is 20 MB. (AT01)"))
        f = self.bot.get_file(file_id)
        if not mime:
            ext = os.path.splitext(f.file_path)[1]
            mime = mimetypes.guess_type(f.file_path, strict=False)[0]
        else:
            ext = mimetypes.guess_extension(mime, strict=False)
        file = tempfile.NamedTemporaryFile(suffix=ext)
        full_path = file.name
        f.download(out=file)
        file.seek(0)
        mime = getattr(file_obj, "mime_type", mime or magic.from_file(full_path, mime=True))
        if type(mime) is bytes:
            mime = mime.decode()
        return file, mime, os.path.basename(full_path), full_path

    def _download_gif(self, file: telegram.File) -> Tuple[IO[bytes], str, str, str]:
        """
        Download and convert GIF image.

        Args:
            file: Telegram File object

        Returns:
            Tuple[IO[bytes], str, str, str]:
                ``tempfile`` file-like object, MIME type, proposed file name
        """
        file, _, filename, path = self._download_file(file, 'video/mpeg')
        gif_file = tempfile.NamedTemporaryFile(suffix='.gif')
        VideoFileClip(path).write_gif(gif_file, program="ffmpeg")
        file.close()
        gif_file.seek(0)
        return gif_file, "image/gif", os.path.basename(gif_file.file_path), gif_file.file_path



# coding=utf-8

from telegram import Update
from telegram.ext import CommandHandler


class GlobalCommandHandler(CommandHandler):
    """Global command handler, handles commands from both chats and channels."""

    def check_update(self, update):
        """Determines whether an update should be passed to this handlers :attr:`callback`.

        Args:
            update (:class:`telegram.Update`): Incoming telegram update.

        Returns:
            :obj:`bool`
        """
        if (isinstance(update, Update)
                and ((update.message or update.edited_message and self.allow_edited) or
                     (update.channel_post or update.edited_channel_post and self.allow_edited))):
            message = update.message or update.edited_message

            if message.text and message.text.startswith('/') and len(message.text) > 1:
                command = message.text[1:].split(None, 1)[0].split('@')
                command.append(
                    message.bot.username)  # in case the command was send without a username

                if self.filters is None:
                    res = True
                elif isinstance(self.filters, list):
                    res = any(func(message) for func in self.filters)
                else:
                    res = self.filters(message)

                return res and (command[0].lower() in self.command
                                and command[1].lower() == message.bot.username.lower())
            else:
                return False

        else:
            return False



# coding=utf-8

from typing import Tuple, Dict, TYPE_CHECKING, List

from telegram import Message, Update
from telegram.ext import CommandHandler, ConversationHandler, RegexHandler, CallbackQueryHandler

from ehforwarderbot import coordinator, EFBChannel
from ehforwarderbot.message import EFBMsgCommand
from .constants import Flags
from .locale_mixin import LocaleMixin

if TYPE_CHECKING:
    from . import TelegramChannel


class ETMCommandMsgStorage:
    def __init__(self, command: List[EFBMsgCommand], channel: EFBChannel, prefix: str, body: str):
        self.commands = command
        self.channels = channel
        self.prefix = prefix
        self.body = body


class CommandsManager(LocaleMixin):
    """
    Functions related to Command messages and
    Additional features of slave channels.
    """

    def __init__(self, channel: 'TelegramChannel'):
        self.channel: 'TelegramChannel' = channel
        self.bot = channel.bot_manager
        self.msg_storage: Dict[Tuple[int, int], ETMCommandMsgStorage] = dict()

        self.bot.dispatcher.add_handler(
            CommandHandler("extra", self.extra_help))
        self.bot.dispatcher.add_handler(
            RegexHandler(r"^/(?P<id>[0-9]+)_(?P<command>[a-z0-9_-]+)", self.extra_call,
                         pass_groupdict=True))

        self.command_conv = ConversationHandler(
            entry_points=[],
            states={Flags.COMMAND_PENDING: [CallbackQueryHandler(self.command_exec)]},
            fallbacks=[CallbackQueryHandler(self.bot.session_expired)],
            per_message=True,
            per_chat=True,
            per_user=False
        )

        self.bot.dispatcher.add_handler(self.command_conv)

    def register_command(self, message: Message, commands: ETMCommandMsgStorage):
        message_identifier = (message.chat.id, message.message_id)
        self.command_conv.conversations[message_identifier] = Flags.COMMAND_PENDING
        self.msg_storage[message_identifier] = commands

    def command_exec(self, bot, update: Update) -> int:
        """
        Run a command from a command message.
        Triggered by callback message with status `Flags.COMMAND_PENDING`.

        This method is a part of the command message conversation handler.

        Args:
            bot: Telegram Bot instance
            update: The update

        Returns:
            The next state
        """

        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id
        callback = update.callback_query.data

        index = (chat_id, message_id)

        if not callback.isdecimal():
            msg = self._("Invalid parameter: {0}. (CE01)").format(callback)
            self.msg_storage.pop(index, None)
            self.bot.edit_message_text(text=msg, chat_id=chat_id, message_id=message_id)
            return ConversationHandler.END
        elif not (0 <= int(callback) < len(self.msg_storage[index].commands)):
            msg = self._("Index out of bound: {0}. (CE02)").format(callback)
            self.msg_storage.pop(index, None)
            self.bot.edit_message_text(text=msg, chat_id=chat_id, message_id=message_id)
            return ConversationHandler.END

        callback = int(callback)
        command_storage = self.msg_storage[index]
        channel_id = command_storage.channels.channel_id
        command = command_storage.commands[callback]
        prefix = "%s\n%s\n--------" % (command_storage.prefix, command_storage.body)

        fn = getattr(coordinator.slaves[channel_id], command.callable_name, None)
        if fn is not None:
            msg = fn(*command.args, **command.kwargs)
        else:
            msg = self._command_fallback(*command.args, __channel_id=channel_id, __callable=command.callable,
                                         **command.kwargs)
        self.msg_storage.pop(index, None)
        self.bot.edit_message_text(prefix=prefix, text=msg,
                                   chat_id=chat_id, message_id=message_id)
        return ConversationHandler.END

    def extra_help(self, bot, update):
        """
        Show list of additional features and their usage.
        Triggered by `/extra`.

        Args:
            bot: Telegram Bot instance
            update: Message update
        """
        msg = self._("List of slave channel features:")
        for n, i in enumerate(sorted(coordinator.slaves)):
            i = coordinator.slaves[i]
            msg += "\n\n<b>%s %s</b>" % (i.channel_emoji, i.channel_name)
            extra_fns = i.get_extra_functions()
            if extra_fns:
                for j in extra_fns:
                    fn_name = "/%s_%s" % (n, j)
                    msg += "\n\n%s <b>(%s)</b>\n%s" % (
                        fn_name, extra_fns[j].name, extra_fns[j].desc.format(function_name=fn_name))
            else:
                msg += self._("\nNo command found.")
        self.bot.send_message(update.effective_chat.id, msg, parse_mode="HTML")

    def extra_call(self, bot, update, groupdict: Dict[str, str] = None):
        """
        Invoke an additional feature from slave channel.

        Args:
            bot: Telegram Bot instance
            update: Message update
            groupdict: Parameters offered by the message
                'id': Index of channel sorted by ``channel_id`` in lexicographical order.
                'command': Name of the command.
        """
        if int(groupdict['id']) >= len(coordinator.slaves):
            return self.bot.reply_error(update, self._("Invalid slave channel ID. (XC01)"))

        slaves = coordinator.slaves

        channel = slaves[sorted(slaves)[int(groupdict['id'])]]
        functions = channel.get_extra_functions()

        if groupdict['command'] not in functions:
            return self.bot.reply_error(update, self._("Command not found in selected channel. (XC02)"))

        header = "%s %s: %s\n-------\n" % (
            channel.channel_emoji, channel.channel_name, functions[groupdict['command']].name)
        msg = self.bot.send_message(update.message.chat.id,
                                    prefix=header, text=self._("Please wait..."))

        result = functions[groupdict['command']](" ".join(update.message.text.split(' ', 1)[1:]))

        self.bot.edit_message_text(prefix=header, text=result,
                                   chat_id=update.message.chat.id, message_id=msg.message_id)

    def _command_fallback(self, *args, __channel_id: str, __callable: str, **kwargs) -> str:
        return self._("Error: Command is not found in the channel.\n"
                      "Function: {channel_id}.{callable}\n"
                      "Arguments: {args!r}\nKeyword Arguments: {kwargs!r}").format(channel_id=__channel_id,
                                                                                   callable=__callable,
                                                                                   args=args,
                                                                                   kwargs=kwargs)



# coding=utf-8

import html
import logging
import os
import tempfile
import traceback
import urllib.parse
from typing import Tuple, Optional, TYPE_CHECKING, List

import pydub
import telegram
import telegram.constants
import telegram.error
import telegram.ext

from ehforwarderbot import EFBMsg, EFBStatus, coordinator
from ehforwarderbot.constants import MsgType, ChatType
from ehforwarderbot.exceptions import EFBMessageError
from ehforwarderbot.message import EFBMsgLinkAttribute, EFBMsgLocationAttribute, EFBMsgCommand
from ehforwarderbot.status import EFBChatUpdates, EFBMemberUpdates, EFBMessageRemoval
from . import utils, ETMChat
from .commands import ETMCommandMsgStorage
from .constants import Emoji
from .locale_mixin import LocaleMixin

if TYPE_CHECKING:
    from . import TelegramChannel
    from .bot_manager import TelegramBotManager
    from .db import DatabaseManager


class SlaveMessageProcessor(LocaleMixin):
    """Process messages as EFBMsg objects from slave channels."""

    def __init__(self, channel: 'TelegramChannel'):
        self.channel: 'TelegramChannel' = channel
        self.bot: 'TelegramBotManager' = self.channel.bot_manager
        self.logger: logging.Logger = logging.getLogger(__name__)
        self.flag: utils.ExperimentalFlagsManager = self.channel.flag
        self.db: 'DatabaseManager' = channel.db

    def send_message(self, msg: EFBMsg) -> EFBMsg:
        """
        Process a message from slave channel and deliver it to the user.

        Args:
            msg (EFBMsg): The message.
        """
        try:
            xid = msg.uid
            self.logger.debug("[%s] Slave message delivered to ETM.\n%s", xid, msg)

            chat_uid = utils.chat_id_to_str(chat=msg.chat)
            tg_chat = self.db.get_chat_assoc(slave_uid=chat_uid)
            if tg_chat:
                tg_chat = tg_chat[0]

            self.logger.debug("[%s] The message should deliver to %s", xid, tg_chat)

            if tg_chat == ETMChat.MUTE_CHAT_ID:
                self.logger.debug("[%s] Sender of the message is muted.", xid)
                return msg

            multi_slaves = False

            if tg_chat:
                slaves = self.db.get_chat_assoc(master_uid=tg_chat)
                if slaves and len(slaves) > 1:
                    multi_slaves = True
                    self.logger.debug("[%s] Sender is linked with other chats in a Telegram group.", xid)

            self.logger.debug("[%s] Message is in chat %s", xid, msg.chat)

            # Generate chat text template & Decide type target
            tg_dest = self.channel.config['admins'][0]
            if tg_chat:  # if this chat is linked
                tg_dest = int(utils.chat_id_str_to_id(tg_chat)[1])

            msg_template = self.generate_message_template(msg, tg_chat, multi_slaves)

            self.logger.debug("[%s] Message is sent to Telegram chat %s, with header \"%s\".",
                              xid, tg_dest, msg_template)

            # When editing message
            old_msg_id: Tuple[str, str] = None
            if msg.edit:
                old_msg = self.db.get_msg_log(slave_msg_id=msg.uid,
                                              slave_origin_uid=utils.chat_id_to_str(chat=msg.chat))
                if old_msg:
                    old_msg_id: Tuple[str, str] = utils.message_id_str_to_id(old_msg.master_msg_id)
                else:
                    self.logger.info('[%s] Was supposed to edit this message, '
                                     'but it does not exist in database. Sending new message instead.',
                                     msg.uid)

            # When targeting a message (reply to)
            target_msg_id: Tuple[str, str] = None
            if isinstance(msg.target, EFBMsg):
                self.logger.debug("[%s] Message is replying to %s.", msg.uid, msg.target)
                log = self.db.get_msg_log(
                    slave_msg_id=msg.target.uid,
                    slave_origin_uid=utils.chat_id_to_str(chat=msg.target.chat)
                )
                if not log:
                    self.logger.debug("[%s] Target message %s is not found in database.", msg.uid, msg.target)
                else:
                    self.logger.debug("[%s] Target message has database entry: %s.", msg.uid, log)
                    target_msg_id = utils.message_id_str_to_id(log.master_msg_id)
                    if not target_msg_id or target_msg_id[0] != str(tg_dest):
                        self.logger.error('[%s] Trying to reply to a message not from this chat. '
                                          'Message destination: %s. Target message: %s.',
                                          msg.uid, tg_dest, target_msg_id)
                    else:
                        target_msg_id = target_msg_id[1]

            commands: Optional[List[EFBMsgCommand]] = None
            reply_markup: Optional[telegram.InlineKeyboardMarkup] = None
            if msg.commands:
                commands = msg.commands.commands
                if old_msg_id:
                    raise EFBMessageError(self._('Command message cannot be edited'))
                buttons = []
                for i, ival in enumerate(commands):
                    buttons.append([telegram.InlineKeyboardButton(ival.name, callback_data=str(i))])
                reply_markup = telegram.InlineKeyboardMarkup(buttons)

            msg.text = msg.text or ""

            # Type dispatching
            if msg.type == MsgType.Text:
                tg_msg = self.slave_message_text(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type == MsgType.Link:
                tg_msg = self.slave_message_link(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type in [MsgType.Image, MsgType.Sticker]:
                tg_msg = self.slave_message_image(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type == MsgType.File:
                tg_msg = self.slave_message_file(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type == MsgType.Audio:
                tg_msg = self.slave_message_audio(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type == MsgType.Location:
                tg_msg = self.slave_message_location(msg, tg_dest, msg_template, old_msg_id, target_msg_id,
                                                     reply_markup)
            elif msg.type == MsgType.Video:
                tg_msg = self.slave_message_video(msg, tg_dest, msg_template, old_msg_id, target_msg_id, reply_markup)
            elif msg.type == MsgType.Unsupported:
                tg_msg = self.slave_message_unsupported(msg, tg_dest, msg_template, old_msg_id, target_msg_id,
                                                        reply_markup)
            else:
                self.bot.send_chat_action(tg_dest, telegram.ChatAction.TYPING)
                tg_msg = self.bot.send_message(tg_dest, prefix=msg_template,
                                               text=self._("Unsupported type of message. (UT01)"))

            if tg_msg and msg.commands:
                self.channel.commands.register_command(tg_msg, ETMCommandMsgStorage(
                    commands, coordinator.slaves[msg.chat.channel_id], msg_template, msg.text
                ))

            self.logger.debug("[%s] Message is sent to the user.", xid)
            if not msg.author.is_system:
                msg_log = {"master_msg_id": utils.message_id_to_str(tg_msg.chat.id, tg_msg.message_id),
                           "text": msg.text or "Sent a %s." % msg.type,
                           "msg_type": msg.type,
                           "sent_to": "master" if msg.author.is_self else 'slave',
                           "slave_origin_uid": utils.chat_id_to_str(chat=msg.chat),
                           "slave_origin_display_name": msg.chat.chat_alias,
                           "slave_member_uid": msg.author.chat_uid if not msg.author.is_self else None,
                           "slave_member_display_name": msg.author.chat_alias if not msg.author.is_self else None,
                           "slave_message_id": msg.uid,
                           "update": msg.edit
                           }
                self.db.add_msg_log(**msg_log)
                self.logger.debug("[%s] Message inserted/updated to the database.", xid)
        except Exception as e:
            self.logger.error("[%s] Error occurred while processing message from slave channel.\nMessage: %s\n%s\n%s",
                              xid, repr(msg), repr(e), traceback.format_exc())

    def slave_message_text(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                           old_msg_id: Optional[Tuple[str, str]] = None,
                           target_msg_id: Optional[str] = None,
                           reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        """
        Send message as text to Telegram.
        
        Args:
            msg (EFBMsg): Message
            tg_dest (str): Telegram Chat ID
            msg_template (str): Header of the message
            old_msg_id: Telegram message ID to edit
            target_msg_id: Telegram message ID to reply to
            reply_markup: Reply markup to be added to the message

        Returns:
            The telegram bot message object sent
        """
        self.logger.debug("[%s] Sending as a text message.", msg.uid)
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.TYPING)

        # Join message is Deprecated from ETM v2.0.0a1
        #
        # join_msg_threshold_secs = self._flag('join_msg_threshold_secs', 15)
        #
        # Check if this message should append the previous one
        # Logic:
        #   1. Only append if the message is sent to linked chats
        #   2. Only append if flag ``join_msg_threshold_secs`` > 0
        #   3. Check for the previous non-system message in the Telegram chat sent by the bot
        #      Link if it is also a text message, sent by the same person from the same slave chat,
        #      within specified number of seconds away from this one.
        #
        # if tg_chat_assoced and join_msg_threshold_secs > 0:
        #     last_msg = self.db.get_last_msg_from_chat(tg_dest)
        #     if last_msg:
        #         if last_msg.msg_type == "Text":
        #             append_last_msg = str(last_msg.slave_origin_uid) == "%s.%s" % \
        #                               (msg.channel_id, msg.origin['uid']) \
        #                               and str(last_msg.master_msg_id).startswith(str(tg_dest) + ".") \
        #                               and last_msg.sent_to == "master"
        #             if msg.source == ChatType.Group:
        #                 append_last_msg &= str(last_msg.slave_member_uid) == str(msg.member['uid'])
        #             append_last_msg &= datetime.datetime.now() - last_msg.time <= datetime.timedelta(
        #                 seconds=join_msg_threshold_secs)
        #         else:
        #             append_last_msg = False
        #     else:
        #         append_last_msg = False
        #
        # if append_last_msg:
        #     self.logger.debug("[%s] Appending this message to the previous one.", msg.uid)
        #
        # if tg_chat_assoced and append_last_msg:
        #     self.logger.debug("[%s] Edit telegram message id: %s.\nPrevious message: %s",
        #                       msg.uid, last_msg.master_msg_id, last_msg.text)
        #     msg.text = "%s\n%s" % (last_msg.text, msg.text)
        #
        #     tg_msg = self.bot_edit_message_text(chat_id=tg_dest,
        #                                         message_id=last_msg.master_msg_id.split(".", 1)[1],
        #                                         text=msg.text, prefix=msg_template,
        #                                         parse_mode=parse_mode)
        # else:
        #     self.logger.debug("[%s] Sending as a new text message.", msg.uid)
        #     tg_msg = self.bot_send_message(tg_dest,
        #                                    text=msg.text, prefix=msg_template,
        #                                    parse_mode=parse_mode)
        #     self.logger.debug("[%s] New message is sent to Telegram:\n%s", msg.uid, tg_msg)
        #
        # self.logger.debug("[%s] Message is successfully processed as text message", msg.uid)
        # return tg_msg, append_last_msg

        parse_mode = "HTML"

        if msg.substitutions:
            ranges = sorted(msg.substitutions.keys())
            text = ""
            prev = 0
            for i in ranges:
                text += html.escape(msg.text[prev:i[0]])
                if msg.substitutions[i].is_self:
                    text += '<a href="tg://user?id=%s">' % self.channel.config['admins'][0]
                    text += html.escape(msg.text[i[0]:i[1]])
                    text += "</a>"
                else:
                    text += html.escape(msg.text[i[0]:i[1]])
                prev = i[1]
            text += html.escape(msg.text[prev:])
            msg.text = text
        elif msg.text:
            msg.text = html.escape(msg.text)

        if not old_msg_id:
            tg_msg = self.bot.send_message(tg_dest,
                                           text=msg.text, prefix=msg_template,
                                           parse_mode=parse_mode,
                                           reply_to_message_id=target_msg_id,
                                           reply_markup=reply_markup)
        else:
            # Cannot change reply_to_message_id when editing a message
            tg_msg = self.bot.edit_message_text(chat_id=old_msg_id[0],
                                                message_id=old_msg_id[1],
                                                text=msg.text, prefix=msg_template,
                                                parse_mode=parse_mode,
                                                reply_markup=reply_markup)

        self.logger.debug("[%s] Processed and sent as text message", msg.uid)
        return tg_msg

    def slave_message_link(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                           old_msg_id: Optional[Tuple[str, str]] = None,
                           target_msg_id: Optional[str] = None,
                           reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.TYPING)

        attributes: EFBMsgLinkAttribute = msg.attributes

        thumbnail = urllib.parse.quote(attributes.image or "", safe="?=&#:/")
        thumbnail = "<a href=\"%s\">🔗</a>" % thumbnail if thumbnail else "🔗"
        text = "%s <a href=\"%s\">%s</a>\n%s" % \
               (thumbnail,
                urllib.parse.quote(attributes.url, safe="?=&#:/"),
                html.escape(attributes.title or attributes.url),
                html.escape(attributes.description or ""))
        if msg.text:
            text += "\n\n" + msg.text
        if old_msg_id:
            return self.bot.edit_message_text(text, chat_id=old_msg_id[0], message_id=old_msg_id[1],
                                              prefix=msg_template, parse_mode='HTML',
                                              reply_markup=reply_markup)
        else:
            return self.bot.send_message(chat_id=tg_dest,
                                         text=text,
                                         prefix=msg_template,
                                         parse_mode="HTML",
                                         reply_to_message_id=target_msg_id,
                                         reply_markup=reply_markup)

    def slave_message_image(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                            old_msg_id: Optional[Tuple[str, str]] = None,
                            target_msg_id: Optional[str] = None,
                            reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.UPLOAD_PHOTO)
        self.logger.debug("[%s] Message is of %s type.\nPath: %s\nMIME: %s", msg.uid, msg.type, msg.path, msg.mime)
        self.logger.debug("[%s] Size of %s is %s.", msg.uid, msg.path, os.stat(msg.path).st_size)

        if not msg.text:
            if msg.type == MsgType.Image:
                msg.text = "sent a picture."
            elif msg.type == MsgType.Sticker:
                msg.text = "sent a sticker."
        try:
            if old_msg_id:
                return self.bot.edit_message_caption(chat=old_msg_id[0], message_id=old_msg_id[1],
                                                     prefix=msg_template, caption=msg.text)
            elif msg.mime == "image/gif":
                return self.bot.send_document(tg_dest, msg.file, prefix=msg_template, caption=msg.text,
                                              reply_to_message_id=target_msg_id,
                                              reply_markup=reply_markup)
            else:
                try:
                    return self.bot.send_photo(tg_dest, msg.file, prefix=msg_template, caption=msg.text,
                                               reply_to_message_id=target_msg_id,
                                               reply_markup=reply_markup)
                except telegram.error.BadRequest as e:
                    self.logger.error('[%s] Failed to send it as image, sending as document. Reason: %s', e)
                    return self.bot.send_document(tg_dest, msg.file, prefix=msg_template,
                                                  caption=msg.text, filename=msg.filename,
                                                  reply_to_message_id=target_msg_id,
                                                  reply_markup=reply_markup)
        finally:
            msg.file.close()

    def slave_message_file(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                           old_msg_id: Optional[Tuple[str, str]] = None,
                           target_msg_id: Optional[str] = None,
                           reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.UPLOAD_DOCUMENT)
        if not msg.filename:
            file_name = os.path.basename(msg.path)
            msg.text = "sent a file."
        else:
            file_name = msg.filename
        try:
            if old_msg_id:
                return self.bot.edit_message_caption(chat=old_msg_id[0], message_id=old_msg_id[1],
                                                     prefix=msg_template, caption=msg.text)
            return self.bot.send_document(tg_dest, msg.file,
                                          prefix=msg_template,
                                          caption=msg.text, filename=file_name,
                                          reply_to_message_id=target_msg_id,
                                          reply_markup=reply_markup)
        finally:
            msg.file.close()

    def slave_message_audio(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                            old_msg_id: Optional[Tuple[str, str]] = None,
                            target_msg_id: Optional[str] = None,
                            reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.RECORD_AUDIO)
        msg.text = msg.text or ''
        self.logger.debug("[%s] Message is an audio file.", msg.uid)
        no_conversion = self.flag("no_conversion")
        try:
            if old_msg_id:
                return self.bot.edit_message_caption(chat=old_msg_id[0], message_id=old_msg_id[1],
                                                     prefix=msg_template, caption=msg.text)
            if no_conversion:
                self.logger.debug('[%s] This audio file is sent as a document without converting to OPUS.', msg.uid)
                self.logger.debug("[%s] MIME type reported by the message: %s", msg.uid, msg.mime)
                if msg.mime == "audio/mpeg":
                    tg_msg = self.bot.send_audio(tg_dest, msg.file, prefix=msg_template, caption=msg.text,
                                                 reply_to_message_id=target_msg_id, reply_markup=reply_markup)
                else:
                    tg_msg = self.bot.send_document(tg_dest, msg.file, prefix=msg_template, caption=msg.text,
                                                    reply_to_message_id=target_msg_id, reply_markup=reply_markup)
            else:
                with tempfile.NamedTemporaryFile() as f:
                    pydub.AudioSegment.from_file(msg.file).export(f, format="ogg", codec="libopus",
                                                                  parameters=['-vbr', 'on'])
                    tg_msg = self.bot.send_voice(tg_dest, f, prefix=msg_template, caption=msg.text,
                                                 reply_to_message_id=target_msg_id, reply_markup=reply_markup)
            return tg_msg
        finally:
            msg.file.close()

    def slave_message_location(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                               old_msg_id: Optional[Tuple[str, str]] = None,
                               target_msg_id: Optional[str] = None,
                               reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.FIND_LOCATION)
        attributes: EFBMsgLocationAttribute = msg.attributes
        self.logger.info("[%s] Sending as a Telegram venue.\nlat: %s, long: %s\ntitle: %s\naddress: %s",
                         msg.uid,
                         attributes.latitude, attributes.longitude,
                         msg.text, msg_template)
        if old_msg_id and old_msg_id[0] == tg_dest:
            # TRANSLATORS: Flag for edited message, but cannot be edited on Telegram.
            msg_template += self._('[edited]')
            target_msg_id = target_msg_id or old_msg_id[1]
        return self.bot.send_venue(tg_dest, latitude=attributes.latitude,
                                   longitude=attributes.longitude, title=msg.text,
                                   address=msg_template, reply_to_message_id=target_msg_id,
                                   reply_markup=reply_markup)

    def slave_message_video(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                            old_msg_id: Optional[Tuple[str, str]] = None,
                            target_msg_id: Optional[str] = None,
                            reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.UPLOAD_VIDEO)
        if not msg.text:
            msg.text = "sent a video."
        try:
            if old_msg_id:
                return self.bot.edit_message_caption(chat=old_msg_id[0], message_id=old_msg_id[1],
                                                     prefix=msg_template, caption=msg.text)
            return self.bot.send_video(tg_dest, msg.file, prefix=msg_template, caption=msg.text,
                                       reply_to_message_id=target_msg_id,
                                       reply_markup=reply_markup)
        finally:
            msg.file.close()

    def slave_message_unsupported(self, msg: EFBMsg, tg_dest: str, msg_template: str,
                                  old_msg_id: Optional[Tuple[str, str]] = None,
                                  target_msg_id: Optional[str] = None,
                                  reply_markup: Optional[telegram.ReplyMarkup] = None) -> telegram.Message:
        self.logger.debug("[%s] Sending as an unsupported message.", msg.uid)
        self.bot.send_chat_action(tg_dest, telegram.ChatAction.TYPING)

        if not old_msg_id:
            tg_msg = self.bot.send_message(tg_dest,
                                           text=msg.text, prefix=msg_template + " (unsupported)",
                                           reply_to_message_id=target_msg_id, reply_markup=reply_markup)
        else:
            # Cannot change reply_to_message_id when editing a message
            tg_msg = self.bot.edit_message_text(chat_id=old_msg_id[0],
                                                message_id=old_msg_id[1],
                                                text=msg.text, prefix=msg_template + " (unsupported)",
                                                reply_markup=reply_markup)

        self.logger.debug("[%s] Processed and sent as text message", msg.uid)
        return tg_msg

    def send_status(self, status: EFBStatus):
        if isinstance(status, EFBChatUpdates):
            self.logger.debug("Received chat updates from channel %s", status.channel)
            for i in status.removed_chats:
                self.db.delete_slave_chat_info(status.channel.channel_id, i)
            for i in status.new_chats + status.modified_chats:
                chat = status.channel.get_chat(i)
                self.db.set_slave_chat_info(slave_channel_name=status.channel.channel_name,
                                            slave_channel_emoji=status.channel.channel_emoji,
                                            slave_channel_id=status.channel.channel_id,
                                            slave_chat_name=chat.chat_name,
                                            slave_chat_alias=chat.chat_alias,
                                            slave_chat_type=chat.chat_type,
                                            slave_chat_uid=chat.chat_uid)
        elif isinstance(status, EFBMemberUpdates):
            self.logger.debug("Received member updates from channel %s about group %s",
                              status.channel, status.chat_id)
            self.logger.info('Currently group member info update is ignored.')
        elif isinstance(status, EFBMessageRemoval):
            self.logger.debug("Received message removal request from channel %s on message %s",
                              status.source_channel, status.message)
            old_msg = self.db.get_msg_log(
                slave_msg_id=status.message.uid,
                slave_origin_uid=utils.chat_id_to_str(chat=status.message.chat))
            if old_msg:
                old_msg_id: Tuple[str, str] = utils.message_id_str_to_id(old_msg.master_msg_id)
                try:
                    if not self.channel.flag('prevent_message_removal'):
                        self.bot.delete_message(*old_msg_id)
                        return
                except telegram.TelegramError:
                    pass
                self.bot.send_message(chat_id=old_msg_id[0],
                                      text=self._("Message removed in remote chat."),
                                      reply_to_message_id=old_msg_id[1])
            else:
                self.logger.info('[%s] Was supposed to delete a message, '
                                 'but it does not exist in database: %s', status)

        else:
            self.logger.error('Received an unknown type of update: %s', status)

    def generate_message_template(self, msg: EFBMsg, tg_chat, multi_slaves: bool) -> str:
        msg_prefix = ""  # For group member name
        if msg.chat.chat_type == ChatType.Group:
            self.logger.debug("[%s] Message is from a group. Sender: %s", msg.uid, msg.author)
            msg_prefix = ETMChat(chat=msg.author, db=self.db).display_name

        if tg_chat and not multi_slaves:  # if singly linked
            if msg_prefix:  # if group message
                msg_template = "%s:" % msg_prefix
            else:
                if msg.chat != msg.author:
                    msg_template = "%s:" % ETMChat(chat=msg.author, db=self.db).display_name
                else:
                    msg_template = ""
        elif msg.chat.chat_type == ChatType.User:
            emoji_prefix = msg.chat.channel_emoji + Emoji.get_source_emoji(msg.chat.chat_type)
            name_prefix = ETMChat(chat=msg.chat, db=self.db).display_name
            if msg.chat != msg.author:
                name_prefix += ", %s" % ETMChat(chat=msg.author, db=self.db).display_name
            msg_template = "%s %s:" % (emoji_prefix, name_prefix)
        elif msg.chat.chat_type == ChatType.Group:
            emoji_prefix = msg.chat.channel_emoji + Emoji.get_source_emoji(msg.chat.chat_type)
            name_prefix = ETMChat(chat=msg.chat, db=self.db).display_name
            msg_template = "%s %s [%s]:" % (emoji_prefix, msg_prefix, name_prefix)
        elif msg.chat.chat_type == ChatType.System:
            emoji_prefix = msg.chat.channel_emoji + Emoji.get_source_emoji(msg.chat.chat_type)
            name_prefix = ETMChat(chat=msg.chat, db=self.db).display_name
            msg_template = "%s %s:" % (emoji_prefix, name_prefix)
        else:
            msg_template = self._("Unknown message source ({0}):").format(msg.chat.chat_type)
        return msg_template
